# -*- coding: utf-8 -*-
# pheme/transformation/storage.py
# Copyright (C) 2020-2021 Greenbone AG
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from uuid import uuid4
from typing import Dict

from django.core.cache import cache


def __default_store_handler(name: str, value: Dict):
    cache.set(name, value)


def __default_load_handler(name: str) -> Dict:
    return cache.get(name)


def __default_id_generator(prefix: str) -> str:
    return f"{prefix}-{uuid4()}"


def load(name: str, *, handler=__default_load_handler) -> Dict:
    return handler(name)


def store(
    prefix: str,
    value: Dict,
    *,
    handler=__default_store_handler,
    id_generator=__default_id_generator,
):
    name = id_generator(prefix)
    if isinstance(value, dict):
        value["internal_name"] = name
    handler(name, value)
    return name



# -*- coding: utf-8 -*-
# pheme/views.py
# Copyright (C) 2020-2021 Greenbone AG
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import dataclasses
import rest_framework.renderers
from rest_framework.decorators import api_view, parser_classes, renderer_classes
from rest_framework.response import Response
from rest_framework.request import Request


from pheme.parser.xml import XMLFormParser, XMLParser
from pheme.transformation import scanreport
from pheme.storage import store, load
from pheme.renderer import MarkDownTableRenderer, XMLRenderer, CSVRenderer
from pheme.transformation.scanreport import model
from pheme.version import __version__


@api_view(["GET"])
@renderer_classes([rest_framework.renderers.JSONRenderer])
def load_cache(request, key):
    return Response(load(key))


@api_view(["POST"])
@parser_classes([rest_framework.parsers.JSONParser])
@renderer_classes([rest_framework.renderers.JSONRenderer])
def store_cache(request):
    key = request.data.get("key", "unknown")
    data = request.data.get("value")
    if request.data.get("append"):
        if isinstance(data, dict):
            cached = load(key) or {}
            cached[data.get("name", "unknown")] = data.get("content")
            data = cached
    name = store(key, data, id_generator=str)
    return Response(name)


@api_view(["POST"])
@parser_classes([XMLParser, XMLFormParser])
@renderer_classes([rest_framework.renderers.JSONRenderer])
def transform(request):
    name = store(
        "scanreport",
        dataclasses.asdict(scanreport.gvmd.transform(request.data)),
    )
    return Response(name)


@api_view(["POST"])
@parser_classes([XMLParser])
@renderer_classes([rest_framework.renderers.JSONRenderer])
def unmodified(request):
    return Response(request.data)


@api_view(["GET"])
@renderer_classes([rest_framework.renderers.JSONRenderer])
def template_elements(request: Request, name: str):
    def load_value_of(key) -> str:
        may_val = load(key) or {}
        return may_val

    images = load_value_of(f"{name}images")
    return Response(
        {
            "template": load_value_of(f"{name}html_template"),
            "pdf_css": load_value_of(f"{name}pdf_css"),
            "html_css": load_value_of(f"{name}html_css"),
            "images": images,
        }
    )


@api_view(["GET"])
@renderer_classes(
    [
        rest_framework.renderers.JSONRenderer,
        scanreport.renderer.ReportFormatPDFReport,
        scanreport.renderer.ReportFormatHTMLReport,
        scanreport.renderer.VulnerabilityHTMLReport,
        scanreport.renderer.VulnerabilityPDFReport,
        XMLRenderer,
        CSVRenderer,
    ]
)
def report(request: Request, name: str):
    def load_value_of(key) -> str:
        may_val = load(key) or {}
        return may_val

    if "report_format_editor" in request.accepted_media_type:
        images = load_value_of(f"{name}images")
        return Response(
            {
                "template": load_value_of(f"{name}html_template"),
                "vulnerability_report": load(name),
                "pdf_css": load_value_of(f"{name}pdf_css"),
                "html_css": load_value_of(f"{name}html_css"),
                "images": images,
            }
        )
    data = load(name)
    data["pheme_version"] = int("".join(filter(str.isdigit, __version__)))
    if request.GET.get("without_overview"):
        # remove charts
        data.pop("overview", None)
    if request.META.get("HTTP_ACCEPT") == "application/xml":
        # XML needs exactly one root
        data = {"report": data}
    return Response(data)


@api_view(["GET"])
@renderer_classes(
    [rest_framework.renderers.JSONRenderer, MarkDownTableRenderer]
)
def scanreport_data_description(request):
    return Response(dataclasses.asdict(model.describe()))



# -*- coding: utf-8 -*-
# pheme/transformation/scanreport/renderer.py
# Copyright (C) 2020-2021 Greenbone AG
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
from typing import Dict

from base64 import b64encode

from django.core.cache import cache
from django.template import Template, Context
from rest_framework import renderers
from rest_framework.request import Request
from weasyprint import CSS, HTML
from pheme.settings import DEBUG
from pheme.parameter import load_params
from pheme.authentication import get_username_role
from pheme.errors import TemplateNotFoundError

logger = logging.getLogger(__name__)


def _load_template(name: str, params: Dict = None) -> Template:
    if not params:
        params = load_params()
    templ = params.get(name)
    if not templ:
        raise TemplateNotFoundError(name)
    return Template(templ)


def _enrich(name: str, data: Dict, parameter: Dict) -> Dict:
    data["internal_name"] = name
    return {**parameter, **data}


def _get_request(renderer_context: Dict) -> Request:
    renderer_context = renderer_context or {}  # to throw key error
    return renderer_context["request"]


def _default_not_found_response(
    renderer_context: Dict, request: Request
) -> str:
    resp = renderer_context["response"]
    resp.status_code = 404
    # pylint: disable=W0212
    path = request._request.path
    report_id = path[path.rfind("/") + 1 :]
    return f'"not data found for {report_id}"'


class Report(renderers.BaseRenderer):
    def render(self, data, accepted_media_type=None, renderer_context=None):
        request = _get_request(renderer_context)
        if not data:
            return _default_not_found_response(renderer_context, request)

        name = data.get("internal_name")
        cache_key = f"{self.media_type}/{name}" if name else None
        logger.debug("generating report %s", cache_key)

        if cache_key and not DEBUG:
            cached = cache.get(cache_key)
            if cached:
                return cached
        params = load_params()
        # separate user specific parameter
        user_parameter = params.pop("user_specific", {})
        username, _ = get_username_role(request)
        if username:
            params = {**params, **user_parameter.get(username, {})}
        data["version"] = params.get("version")

        result = self.apply(name, data, params)
        if cache_key:
            cache.set(cache_key, result)
        return result

    def apply(self, name: str, data: Dict, parameter: Dict):
        raise NotImplementedError(
            "Report class requires .apply() to be implemented"
        )


class VulnerabilityHTMLReport(Report):
    __template = "vulnerability_report_html_template"
    __css_template = "vulnerability_report_html_css"
    media_type = "text/html"
    format = "html"

    def apply(self, name: str, data: Dict, parameter: Dict):
        css = _load_template(self.__css_template, parameter).render(
            Context(parameter)
        )
        data["css"] = css

        return _load_template(self.__template).render(
            Context(_enrich(name, data, parameter))
        )


def _replace_inline_svg_with_img_tags(
    html: str, from_index: int = 0, open_tag="<svg ", close_tag="</svg>"
) -> str:
    """
    Is a workaround because WeasyPrint is not capable of dealing with inline svg

    It searches for svg tags and replaces the inline svg within a given
    html document with img containing a base64 encoded data link.

    It is known that we lose the css styling with that hack, but in the moment
    it is considered better than losing the image altogether or trying to
    produce the images before rendering the template.

    Please replace this method as soon as possible with a proper solution.
    """
    # abort condition is when either open_tag or close_tag was not found
    while True:
        from_index = html.find(open_tag, from_index)
        if from_index == -1:
            return html
        to_index = html.find(close_tag, from_index + len(open_tag))
        if to_index == -1:
            return html
        to_index += len(close_tag)
        to_encode = html[from_index:to_index]
        encoded = b64encode(to_encode.encode()).decode()
        img = (
            f'<img src="data:image/svg+xml;charset=utf-8;base64, {encoded}" />'
        )
        html = html[:from_index] + img + html[to_index:]
        from_index = to_index


def enforce_limit(
    data: Dict, parameter: Dict, format_type: str = "pdf"
) -> Dict:
    if not "results" in data:
        return data

    limits = parameter.get("limits", {}).get(format_type, {})
    max_hosts = limits.get("hosts")
    max_results_in_host = limits.get("results")
    logger.debug(
        "limit max_results %s and max_hosts %s", max_results_in_host, max_hosts
    )
    if not max_hosts and not max_results_in_host:
        return data
    host_cut = False
    if max_hosts and len(data["results"]) > max_hosts:
        data["results"] = data["results"][:max_hosts]
        host_cut = True
    result_cut = False
    if max_results_in_host:
        for result in data["results"]:
            if len(result["results"]) > max_results_in_host:
                result["results"] = result["results"][:max_results_in_host]
                result_cut = True
    if host_cut and not result_cut:
        data["comment"] = limits.get(
            "host_limit_msg", f"Host limit: {max_hosts}."
        )
    elif result_cut and not host_cut:
        data["comment"] = limits.get(
            "result_limit_msg", f"Result limit: {max_results_in_host}."
        )
    elif host_cut and result_cut:
        data["comment"] = limits.get(
            "host_result_limit_msg",
            f"Host limit {max_hosts}; Result limit: {max_results_in_host}.",
        )

    return data


class VulnerabilityPDFReport(Report):
    """
    Is used to generate vulnerability reports in PDF.
    It renders html and css templates given data struct and then translate
    the html document to PDF
    """

    __template = "vulnerability_report_pdf_template"
    __css_template = "vulnerability_report_pdf_css"
    media_type = "application/pdf"
    format = "binary"

    def apply(self, name: str, data: Dict, parameter: Dict):
        data = enforce_limit(data, parameter)
        css = _load_template(self.__css_template, parameter).render(
            Context(parameter)
        )
        html_template = _load_template(self.__template, parameter)
        html = html_template.render(Context(_enrich(name, data, parameter)))
        html = _replace_inline_svg_with_img_tags(html)
        logger.debug("created html")
        pdf = HTML(string=html).write_pdf(stylesheets=[CSS(string=css)])
        logger.debug("created pdf")
        return pdf


class ReportFormatHTMLReport(renderers.BaseRenderer):
    media_type = "text/html+report_format_editor"
    format = "html"

    def render(self, data, accepted_media_type=None, renderer_context=None):
        request = _get_request(renderer_context)
        if not data:
            return _default_not_found_response(renderer_context, request)
        template = Template(data["template"])
        data["vulnerability_report"]["css"] = data["html_css"]
        data["vulnerability_report"]["images"] = data.get("images")
        context = Context(data["vulnerability_report"])
        html = template.render(context)
        logger.debug("created html")
        return html


class ReportFormatPDFReport(renderers.BaseRenderer):
    media_type = "application/pdf+report_format_editor"
    format = "binary"

    def render(self, data, accepted_media_type=None, renderer_context=None):
        request = _get_request(renderer_context)
        if not data:
            return _default_not_found_response(renderer_context, request)
        data["vulnerability_report"]["images"] = data.get("images")
        template = Template(data["template"])
        context = Context(data["vulnerability_report"])
        html = template.render(context)
        logger.debug("created html")
        css = Template(data["pdf_css"]).render(context)
        pdf = HTML(string=html).write_pdf(stylesheets=[CSS(string=css)])
        logger.debug("created pdf")
        return pdf



# -*- coding: utf-8 -*-
# pheme/templatetags/charts.py
# Copyright (C) 2020-2021 Greenbone AG
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import math
from typing import Dict
from django.utils.safestring import SafeText
from pheme.templatetags.charts import (
    calculate_legend_start_height,
    register,
    _severity_class_colors,
    build_legend,
)

__PIE_CHART_TEMPLATE = """
<svg width="{size}" viewBox="0 0 {size} {height}" xmlns="http://www.w3.org/2000/svg">
<g transform="translate({start_pie}, 0)">
{donut}
</g>
{legend}
</svg>
"""

__SLICE_TEMPLATE = """
<g>
<circle cx="{cx}" cy="{cy}" r="{radius}" stroke="{color}" stroke-width="{stroke_width}" stroke-dasharray="{d_array}" stroke-dashoffset="{s_offset}" transform="rotate({r_start}, {cx}, {cy})" fill="transparent"></circle>
<text style="font-size:{font_size}px;font-family:{font_family}" x="{t_x}" y="{t_y}" text-anchor="middle">{label}</text>
</g>
"""


@register.filter
def pie_chart(
    input_values: Dict,
    title_color: Dict = None,
    width: int = 390,
    border_size: int = 0,
    slice_width: int = 90,
    font_family: str = "Dejavu Sans",
    font_size: int = 10,
) -> SafeText:
    """
    creates a pie chart svg.

    The values parameter needs to be a dict with a categoryname: numeric_value.
    E.g.:
    {
        "High": 5,
        "Medium": 3,
        "Low": 2
    }

    The keys need to match the title_color keys. As a default the
    severity_classes: High , Medium, Low are getting used.

    Parameters:
        input_values: dict containing the data as label: numeric_value
        title_color: dict containing the data as label: color
        width: int the height and width of the
        border_size: int size of the border of slice
        slice_width: int the width of slice
        font_family - the font family used within text elements
        font_size - the font size used within text elements

    Returns:
    A SafeString with SVG
    """
    if not title_color:
        title_color = _severity_class_colors
    total = sum(input_values.values())
    if total == 0:
        return SafeText("")
    # pylint: disable=C0103
    # we start at 12' o clock
    angle_offset = -90

    max_legend_len = max([len(k) for k in title_color.keys()])
    chart_size = width - max_legend_len * font_size - font_size

    cx = chart_size / 2  # shift x
    cy = chart_size / 2  # shift y
    # need to cut out size of the slice_width, otherwise edges will be cut off
    radius = (chart_size - slice_width) / 2
    circumference = 2 * math.pi * radius
    dash_array = circumference - border_size

    donut = ""
    for category, amount in input_values.items():
        percent = amount / total
        color = title_color.get(category)
        dash_offset = circumference - percent * circumference
        degrees = angle_offset
        t_angle = (percent * 360) / 2 + angle_offset
        t_radians = t_angle * (math.pi / 180)
        t_x = radius * math.cos(t_radians) + cx
        t_y = radius * math.sin(t_radians) + cy

        angle_offset += percent * 360
        donut += __SLICE_TEMPLATE.format(
            cx=cx,
            cy=cy,
            radius=radius,
            color=color,
            stroke_width=slice_width,
            d_array=dash_array,
            s_offset=dash_offset,
            r_start=degrees,
            t_x=t_x,
            t_y=t_y,
            label=f"{amount} ({round(percent * 100)}%)",
            font_family=font_family,
            font_size=font_size,
        ).strip()

    legend_start = calculate_legend_start_height(
        chart_size, title_color, font_size
    )
    legend = build_legend(legend_start, title_color)
    return SafeText(
        __PIE_CHART_TEMPLATE.format(
            start_pie=width - chart_size,
            size=width,
            height=chart_size,
            donut=donut,
            legend=legend,
        ).strip()
    )



# -*- coding: utf-8 -*-
# pheme/settings.py
# Copyright (C) 2020-2021 Greenbone AG
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Django settings for pheme project.

Generated by 'django-admin startproject' using Django 3.1.

For more information on this file, see
https://docs.djangoproject.com/en/3.1/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/3.1/ref/settings/
"""

from pathlib import Path
import secrets
import os

# Build paths inside the project like this: BASE_DIR / 'subdir'.
__configured_base = os.environ.get("PHEME_BASE_PATH")
BASE_DIR = (
    Path(__configured_base)
    if __configured_base
    else Path(__file__).resolve(strict=True).parent.parent
)
STATIC_DIR = BASE_DIR / "static"
TEMPLATE_DIR = BASE_DIR / "template"
# set default to actual gos path instead of static dir

GSAD_URL = os.environ.get("GSAD_URL", "https://localhost/gmp")

SENTRY_DSN = os.environ.get("SENTRY_DSN_PHEME")

# load sentry_sdk only when SENTRY_DSN_PHEME is set otherwise don't bother
if SENTRY_DSN is not None:
    # pylint: disable=import-error
    import sentry_sdk
    from sentry_sdk.integrations.django import DjangoIntegration

    # pylint: disable=E0110
    # Due to the MYPY handling within sentry_sdk.init, to have a nicer user
    # experience than to have to look up lambdas, pylint detects this as an
    # abstract class; which it is.
    sentry_sdk.init(
        SENTRY_DSN,
        traces_sample_rate=1.0,
        environment=os.environ.get("SENTRY_ENVIRONMENT"),
        server_name=os.environ.get("SENTRY_SERVER_NAME"),
        integrations=[DjangoIntegration()],
    )

DATA_UPLOAD_MAX_MEMORY_SIZE = None

PHEME_CONFIGURATION_PATH = Path(
    os.environ.get(
        "PHEME_CONFIGURATION_PATH",
        "/var/lib/pheme" if Path("/var/lib/pheme").exists() else Path("/tmp/"),
    )
)


SECRET_KEY_LOCATION = PHEME_CONFIGURATION_PATH.joinpath("api_key")

PARAMETER_FILE_ADDRESS = PHEME_CONFIGURATION_PATH.joinpath("parameter.json")
DATA_OBJECT_PATH = "/opt/greenbone/feed/gvmd"
GOS_VERSION = "22.04"
DEFAULT_PARAMETER_ADDRESS = (
    os.environ.get("DEFAULT_PARAMETER_FILE_ADDRESS")
    or f"{DATA_OBJECT_PATH}/{GOS_VERSION}/pheme/default-parameter.json"
)


def __load_or_create_api_key() -> str:
    if SECRET_KEY_LOCATION.exists():
        may_token = SECRET_KEY_LOCATION.read_text(encoding="utf-8")
        if len(may_token.strip()) > 0:
            return may_token
    token = secrets.token_urlsafe(50)
    SECRET_KEY_LOCATION.write_text(token, encoding="utf-8")
    return token


SECRET_KEY = os.environ.get("PHEME_API_KEY", __load_or_create_api_key())

ENV_HOSTS = os.environ.get("ALLOWED_HOSTS", "*")
DEBUG = os.environ.get("PHEME_DEBUG", "").lower() == "true"
ALLOWED_HOSTS = ENV_HOSTS.split(" ") if ENV_HOSTS else []

# Application definition

INSTALLED_APPS = [
    "pheme",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "rest_framework",
]

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

ROOT_URLCONF = "pheme.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [TEMPLATE_DIR],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ]
        },
    }
]

WSGI_APPLICATION = "pheme.wsgi.application"


# Database
# https://docs.djangoproject.com/en/3.1/ref/settings/#databases

DATABASES = {}


# Password validation
# https://docs.djangoproject.com/en/3.1/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = []


# Internationalization
# https://docs.djangoproject.com/en/3.1/topics/i18n/

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/3.1/howto/static-files/

STATIC_URL = "/static/"
STATICFILES_DIRS = [STATIC_DIR, TEMPLATE_DIR]

# Logging
# https://docs.djangoproject.com/en/3.1/topics/logging/

LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "handlers": {
        "console": {"class": "logging.StreamHandler", "formatter": "standard"},
        "syslog": {
            "class": "logging.handlers.SysLogHandler",
            "formatter": "standard",
            "facility": "user",
            "address": "/dev/log",
        },
        "file": {
            "class": "logging.FileHandler",
            "formatter": "standard",
            "filename": "/var/log/pheme/pheme.log"
            if Path("/var/log/pheme").exists()
            else "/tmp/pheme.log",
        },
    },
    "formatters": {
        "standard": {
            "format": "{module}#{funcName} {levelname} {asctime}: {message}",
            "style": "{",
        }
    },
    "root": {
        "handlers": [
            "syslog" if Path("/dev/log").exists() and not DEBUG else "console",
            "file",
        ],
        "level": os.getenv("DJANGO_LOG_LEVEL", "INFO"),
    },
}
# https://docs.djangoproject.com/en/3.1/topics/cache/
CACHES = {
    "default": {
        "BACKEND": "django.core.cache.backends.filebased.FileBasedCache",
        "LOCATION": "/tmp/django_cache",
        "TIMEOUT": 1 * 60 * 2 * 60,  # 2 hours
    }
}

# testing
REST_FRAMEWORK = {
    "TEST_REQUEST_RENDERER_CLASSES": [
        "pheme.renderer.XMLRenderer",
        "rest_framework.renderers.JSONRenderer",
        "rest_framework.renderers.MultiPartRenderer",
    ]
}



# -*- coding: utf-8 -*-
# Copyright (C) 2020-2021 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import mimetypes
import logging
from typing import Dict, Callable
from pathlib import Path
import json
import rest_framework
from django.core.files.uploadedfile import UploadedFile
from rest_framework.decorators import (
    api_view,
    parser_classes,
    renderer_classes,
    authentication_classes,
)
from rest_framework.request import HttpRequest
from rest_framework.response import Response
from pheme.datalink import as_datalink
from pheme import settings
import pheme.authentication


logger = logging.getLogger(__name__)


def __load_params(from_path: str) -> Dict:
    """
    either loads json file from given_path or returns an empty dict when file
    does not exist.

    """
    param_file_obj = Path(from_path)
    return (
        json.loads(param_file_obj.read_text(encoding="utf-8"))
        if param_file_obj.exists()
        else {}
    )


def load_params(
    system_parameter_path: str = settings.PARAMETER_FILE_ADDRESS,
    default_parameter_path: str = settings.DEFAULT_PARAMETER_ADDRESS,
) -> Dict:
    """
    loads default and system parameter and combines them so that system
    parameter may override default parameter.

    A system parameter is set and handled by an individual system while default
    parameter usually come from data-objects.
    """
    return {
        **__load_params(default_parameter_path),
        **__load_params(system_parameter_path),
    }


def __store(params: Dict, *, from_path: str = None) -> Dict:
    Path(from_path).write_text(json.dumps(params), encoding="utf-8")
    return params


def __put(
    request: HttpRequest,
    func: Callable[[HttpRequest, Dict], Dict],
    *,
    from_path: str = settings.PARAMETER_FILE_ADDRESS,
    store: Callable[[Dict, str], Dict] = __store,
) -> Response:
    params = __load_params(from_path=from_path)
    username = request.META.get("GVM_USERNAME")
    if username:
        all_user = params.get("user_specific", {})
        specific_user_params = all_user.get(username, {})
        specific_user_params = func(request, specific_user_params)
        all_user[username] = specific_user_params
        params["user_specific"] = all_user
        value = params
    else:
        value = func(request, params)
    return Response(store(value, from_path=from_path))


def __process_form_data(request: HttpRequest, data: Dict) -> Dict:
    if not isinstance(request.data, dict):
        raise TypeError(
            "Request data is expected to be a dict, "
            f"but it is {type(request.data)}"
        )
    for key, value in request.data.items():
        if isinstance(value, UploadedFile):
            file_type, _ = mimetypes.guess_type(value.name)
            logger.info(
                "uploading filetype %s/%s for %s", file_type, value.name, key
            )
            if file_type and file_type.startswith("image"):
                data[key] = as_datalink(value.read(), file_type)
            elif file_type and file_type.startswith("text"):
                data[key] = value.read().decode()
            else:
                raise ValueError("Only image or text is permitted")
        else:
            data[key] = value
    return data


def __process_json_object(request: HttpRequest, data: Dict) -> Dict:
    return {**data, **request.data}


@api_view(["PUT"])
@parser_classes([rest_framework.parsers.JSONParser])
@renderer_classes([rest_framework.renderers.JSONRenderer])
@authentication_classes(
    [
        pheme.authentication.LoggedInAsAUser,
        pheme.authentication.SimpleApiKeyAuthentication,
    ]
)
def put_value(request: HttpRequest, key: str) -> Response:
    def __process_single_value(request: HttpRequest, data: Dict) -> Dict:
        data[key] = request.data
        return data

    return __put(request, __process_single_value)


@api_view(["PUT"])
@parser_classes(
    [rest_framework.parsers.JSONParser, rest_framework.parsers.MultiPartParser]
)
@renderer_classes([rest_framework.renderers.JSONRenderer])
@authentication_classes(
    [
        pheme.authentication.LoggedInAsAUser,
        pheme.authentication.SimpleApiKeyAuthentication,
    ]
)
def put(request: HttpRequest) -> Response:
    if request.content_type == "application/json":
        return __put(request, __process_json_object)
    return __put(request, __process_form_data)



# -*- coding: utf-8 -*-
# pheme/urls.py
# Copyright (C) 2020-2021 Greenbone AG
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""pheme URL Configuration

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/3.1/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  path('', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  path('', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.urls import include, path
    2. Add a URL to urlpatterns:  path('blog/', include('blog.urls'))
"""

from django.urls import path
import pheme.version
import pheme.views
import pheme.parameter

urlpatterns = [
    path(
        "parameter/<str:key>",
        pheme.parameter.put_value,
        name="put_value_parameters",
    ),
    path("parameter", pheme.parameter.put, name="put_parameter"),
    path("cache/<str:key>", pheme.views.load_cache, name="load_cache"),
    path("cache", pheme.views.store_cache, name="store_cache"),
    path("unmodified", pheme.views.unmodified, name="unmodified"),
    path("transform", pheme.views.transform, name="transform"),
    path("transform/", pheme.views.transform),
    path(
        "scanreport/data/description",
        pheme.views.scanreport_data_description,
        name="scanreport_data_description",
    ),
    path("report/<str:name>", pheme.views.report, name="report"),
    path(
        "template/elements/<str:name>",
        pheme.views.template_elements,
        name="template_elements",
    ),
]



# -*- coding: utf-8 -*-
# pheme/templatetags/bar_chart.py
# Copyright (C) 2020-2021 Greenbone AG
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import itertools
from typing import Dict
from PIL import ImageFont
from django.utils.safestring import SafeText
from pheme.templatetags.charts import (
    register,
    _severity_class_colors,
    build_legend,
    calculate_legend_start_height,
)


__ORIENTATION_LINE_TEMPLATE = """
<rect x="{x}" y="0" height="{height}" width="1" style="fill: #000000;"></rect>
"""

__ORIENTATION_LINE_TEXT_TEMPLATE = """
<text class="orientation label" y="22" x="{x}" fill="#4C4C4D" dominant-baseline="central"
style="font-size:{font_size}px;font-family:{font_family};text-anchor: middle;" width="{width}">{label}
</text>
"""

__X_AXIS_LABEL = """
<text class="x axis label" y="{y}" x="{x}" fill="#000000"
style="font-size:{font_size}px;font-family:{font_family};text-anchor: middle;">{axis_label}
</text>
"""

__BAR_ELEMENT_TEMPLATE = """
<rect x="{x}" y="17" height="10" width="{width}" style="fill: {color};"></rect>
"""

__BAR_TEMPLATE = """
<g class="entry" transform="translate(0, {y})">
<text class="label category" y="22" x="87.5" fill="#4C4C4D" dominant-baseline="central"
style="font-size:{font_size}px;font-family:{font_family};text-anchor: right;" width="{max_hostname_len}">{key}
</text>
<g transform="translate({max_hostname_len}, 0)">
{orientation_lines}
{bar_elements}
</g>
<g transform="translate({begin_total}, 0)">
<text class="sum" y="22" x="10" fill="#4C4C4D" dominant-baseline="central"
style="font-size:{font_size}px;font-family:{font_family};text-anchor: left;" width="100">{total}</text>
</g>
</g>
"""
__BAR_CHART_TEMPLATE = """
<svg width="{width}" viewBox="0 0 {width} {height}" xmlns="http://www.w3.org/2000/svg">
{bars}
<g transform="translate(0, {bar_legend_y})">
<text 
    class="label category" 
    y="22" x="87.5" 
    fill="#4C4C4D" dominant-baseline="central"
    style="font-size:{font_size}px;font-family:{font_family};text-anchor: right;" 
    width="{max_hostname_len}">
    {x_title}
    </text>
</g>
<g transform="translate({max_hostname_len}, {bar_legend_y})">
{bar_legend}
</g>
{legend}
{x_axis_label}
</svg>
"""


@register.filter
@register.simple_tag
def h_bar_chart(
    chart_data: Dict[str, Dict[str, int]],
    x_title: str = "",
    title_color=None,
    svg_width=800,
    bar_jump=44,
    orientation_marker=6,
    limit=10,
    font_family="Dejavu Sans",
    font_file_name="DejaVuSans.ttf",
    font_size=10,
) -> SafeText:
    """
    Returns a stacked horizontal bar chart in svg.

    In order to function it needs a dict with label as key containing a dict
    with color key and amount.

    An example for valid input is:

    {
      "192.168.123.52": {
        "High": 3,
        "Medium": 3,
        "Low": 1
      }
    }

    parameter:
        chart_data - The dict containing the values for the chart
        title_color - A colortable, defaulting to __severity_class_colors
        svg_width - The overall width of the output chart, default 800
        bar_jump - The amount of y pixels to jump from bar element to the next.
            Default is 44.
        orientation_marker - if higher than 0 than there will be a vertical line
            each nth step.
        limit - limits the data by N first elements
        font_family - the font family used within text elements
        font_size - the font size used within text elements
        x_title - the title of the x axis
    """
    data = dict(itertools.islice(chart_data.items(), limit))
    if not title_color:
        title_color = _severity_class_colors
    if not data.values():
        return SafeText("")
    try:
        font = ImageFont.truetype(font_file_name, font_size)
        # add 87.5 + 10 for legend
        max_hostname_len = (
            max(
                max(font.getlength(k) for k in data.keys()),
                font.getlength(x_title),
            )
            + 87.5
            + 10
        )
    except OSError:
        # multiply by 1.25 for kerning and add 87.5 for legend
        max_hostname_len = (
            max(max(len(k) for k in data.keys()), len(x_title))
            * font_size
            * 1.25
            + 87.5
        )
    max_width = svg_width - max_hostname_len - 100  # key and total placeholder
    # highest sum of counts
    max_sum = max([sum(list(counts.values())) for counts in data.values()])
    if max_sum == 0:
        max_sum = 1

    orientation_lines = ""
    orientation_labels = ""
    orientation_basis = (
        int(max_sum / orientation_marker) if orientation_marker > 0 else 0
    )
    if max_sum > orientation_basis and orientation_basis > 0:
        overhead = max_sum % orientation_basis
        # if it's already adjusted we don't need to add anything
        if overhead == 0:
            overhead = orientation_basis
        max_sum = max_sum + orientation_basis - overhead

        for i in range(int(max_sum / orientation_basis + 1)):
            x_pos = i * orientation_basis / max_sum * max_width
            label = str(i * orientation_basis)
            orientation_lines += __ORIENTATION_LINE_TEMPLATE.format(
                x=x_pos,
                height=bar_jump + 10,
                font_family=font_family,
                font_size=font_size,
            )
            orientation_labels += __ORIENTATION_LINE_TEXT_TEMPLATE.format(
                x=x_pos,
                width=len(label),
                label=label,
                font_family=font_family,
                font_size=font_size,
            )

    bars = ""
    for i, (key, counts) in enumerate(data.items()):
        element_x = 0
        elements = ""
        for category, count in counts.items():
            color = title_color.get(category)
            width = count / max_sum * max_width
            elements += __BAR_ELEMENT_TEMPLATE.format(
                x=element_x,
                width=width,
                color=color,
                font_family=font_family,
                font_size=font_size,
                max_hostname_len=max_hostname_len,
            )
            element_x += width

        bars += __BAR_TEMPLATE.format(
            y=i * bar_jump,
            key=key,
            bar_elements=elements,
            orientation_lines=orientation_lines,
            total=sum(counts.values()),
            font_family=font_family,
            font_size=font_size,
            max_hostname_len=max_hostname_len,
            begin_total=svg_width - 100,
        )
    svg_element_lengths = len(data.keys()) * bar_jump + 60 + font_size
    legend_start = calculate_legend_start_height(
        svg_element_lengths, title_color, font_size
    )
    svg_chart = __BAR_CHART_TEMPLATE.format(
        width=svg_width,
        height=svg_element_lengths,
        bars=bars,
        bar_legend=orientation_labels,
        bar_legend_y=len(data.keys()) * bar_jump + 20,
        legend=build_legend(legend_start, title_color),
        font_family=font_family,
        font_size=font_size,
        max_hostname_len=max_hostname_len,
        x_axis_label=__X_AXIS_LABEL.format(
            axis_label="Number of Vulnerabilities",
            font_size=font_size,
            font_family=font_family,
            x=svg_width / 2,
            y=svg_element_lengths,
        ),
        x_title=x_title,
    )
    return SafeText(svg_chart)



# -*- coding: utf-8 -*-
# pheme/transformation/scanreport/gvmd.py
# Copyright (C) 2020-2021 Greenbone AG
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
pheme.transformation.scanreport.gvmd

contrains the function

- transform

it is a specialized module for gvmd scanreports.
"""
import logging
import time

from typing import Dict, List, Optional


from pheme.transformation.scanreport.model import (
    # Equipment,
    Overview,
    Report,
)

logger = logging.getLogger(__name__)

__threats = ["High", "Medium", "Low"]
__threat_index_lookup = {v: i for i, v in enumerate(__threats)}


def measure_time(func):
    def measure(*args, **kwargs):
        startt = time.process_time()
        result = func(*args, **kwargs)
        logger.info("%s took %s", func.__name__, time.process_time() - startt)
        return result

    return measure


def __tansform_tags(item) -> Optional[Dict[str, str]]:
    if isinstance(item, str):
        split = [i.split("=") for i in item.split("|")]
        return {i[0]: i[1].replace("\n", " ") for i in split if len(i) == 2}
    return None


def __group_refs(refs: Dict[str, str]) -> Dict:
    refs_ref = {}
    for ref in refs.get("ref", []):
        if isinstance(ref, dict):
            typus = ref.get("type", "unknown")
            refs_ref[typus] = refs_ref.get(typus, []) + [ref.get("id")]
    return refs_ref


def __get_hostname_from_result(result) -> str:
    if isinstance(result, str):
        return result
    if isinstance(result, dict):
        host = result.get("host", {})
        if isinstance(host, dict):
            return host.get("text", "unknown")
        return host
    return "unknown"


def __return_highest_threat(threats: List[int]) -> str:
    """
    returns the highest threat
    """
    for i, value in enumerate(threats):
        if value > 0:
            return __threats[i]
    return "NA"


def __host_threat_overview(threat_count: Dict) -> Dict:
    """
    returns nvt statistics mostly used in the per host overview
    """
    result = {**threat_count}
    result["total"] = sum(threat_count.values())
    result["highest"] = __return_highest_threat(threat_count.values())
    return result


def __return_highest_severity(severities: List[int]) -> str:
    """
    returns the highest severity within a list
    """
    for i in range(len(severities) - 1, -1, -1):
        if severities[i] > 0:
            return str(i + 1)
    return "NA"


def __host_severity_overview(nvt_count: List[int]) -> Dict:
    """
    returns nvt severity statistics mostly used in the per host overview
    """
    result = {str(i + 1): value for i, value in enumerate(nvt_count)}
    result["total"] = sum(nvt_count)
    result["highest"] = __return_highest_severity(nvt_count)
    return result


def __create_host_information_lookup(report: Dict) -> Dict:
    """
    created a lookup table for available host information
    """
    # lookup for host information name and dict name
    information_key = {"best_os_txt": "os"}

    def filter_per_host(host: Dict) -> Dict:
        information = {}
        found = 0

        def check_host_detail(detail: Dict) -> int:
            name = detail.get("name", "")
            if name in information_key:
                information[information_key.get(name)] = detail.get("value")
                return 1
            return 0

        details = host.get("detail", [])
        if isinstance(details, dict):
            check_host_detail(details)
            return information
        elif isinstance(details, list):
            for detail in details:
                found += check_host_detail(detail)
                if found == len(information_key):
                    return information
            return information
        else:
            return information

    result = {}
    hosts = report.get("host")
    if isinstance(hosts, dict):
        result[hosts.get("ip", "unknown")] = filter_per_host(hosts)
    elif isinstance(hosts, list):
        # best_os_txt seems to be in the end
        for host in hosts[::-1]:
            result[host.get("ip", "unknown")] = filter_per_host(host)
    return result


@measure_time
def __create_results_per_host(report: Dict) -> List[Dict]:
    """
    creates the results dict used by a vulnerability-report based on a given
    gvmd report.
    """
    host_information_lookup = __create_host_information_lookup(report)
    results = report.get("results", {}).get("result", [])
    by_host = {}
    host_threat_count = {}
    host_severity_count = {}
    threat_count = [0] * len(__threats)

    def transform_key(prefix: str, vic: Dict) -> Dict:
        return {f"{prefix}_{key}": value for key, value in vic.items()}

    def per_result(result):
        hostname = __get_hostname_from_result(result)
        host_dict = by_host.get(hostname, {})
        threat = result.get("threat", "unknown")
        port = result.get("port")
        nvt = transform_key("nvt", result.get("nvt", {}))
        nvt["nvt_tags_interpreted"] = __tansform_tags(nvt.get("nvt_tags", ""))
        nvt["nvt_refs_ref"] = __group_refs(nvt.get("nvt_refs", {}))
        qod = transform_key("qod", result.get("qod", {}))
        severity = float(result.get("severity", "0.0"))
        notes = result.get("notes", {}).get("note", [])
        overrides = result.get("overrides", {}).get("override", [])

        new_host_result = {
            "port": port,
            "threat": threat,
            "severity": severity,
            "description": result.get("description"),
            **nvt,
            **qod,
            "notes": [],
            "overrides": [],
        }
        host_results = host_dict.get("results", [])
        host_results.append(new_host_result)
        equipment = host_dict.get("equipment", {})
        ports = list(dict.fromkeys(equipment.get("ports", [])))
        if port and not port.startswith("general"):
            ports = set(ports + [port])
        equipment["ports"] = ports
        if not equipment.get("os"):
            equipment["os"] = host_information_lookup.get(hostname, {}).get(
                "os", "unknown"
            )

        # needs hostname, high, medium, low
        host_threats = host_threat_count.get(
            hostname, {threat: 0 for threat in __threats}
        )
        threat_index = __threat_index_lookup.get(threat)
        if threat_index is not None:
            threat_count[threat_index] += 1
        if host_threats.get(threat) is not None:
            host_threats[threat] += 1
        host_threat_count[hostname] = host_threats

        # needs high, medium, low

        # severity 1 to 10
        host_severities = host_severity_count.get(hostname, [0] * 10)
        if severity > 0:
            host_severities[int(severity) - 1] += 1
        host_severity_count[hostname] = host_severities

        def per_note(note):
            result_notes = new_host_result.get("notes", [])
            new_note = {
                "text": note["text"]["text"],
                "text_excerpt": note["text"]["excerpt"],
            }
            result_notes.append(new_note)

        def per_override(override):
            result_overrides = new_host_result.get("overrides", [])
            severity = override["severity"]
            severity_description = severity
            if not severity:
                severity_description = "Any severity"
            elif float(severity) == 0.0:
                severity_description = "Log"
            elif float(severity) > 0.0:
                severity_description = "Any positive severity"

            new_override = {
                "text": override["text"]["text"],
                "text_excerpt": override["text"]["excerpt"],
                "severity": severity,
                "severity_description": severity_description,
                "new_severity": override["new_severity"],
            }
            result_overrides.append(new_override)

        # lists with just one element can be parsed as dict by xmltodict
        if isinstance(notes, dict):
            per_note(notes)
        else:
            for note in notes:
                per_note(note)

        if isinstance(overrides, dict):
            per_override(overrides)
        else:
            for override in overrides:
                per_override(override)

        by_host[hostname] = {
            "host": hostname,
            "threats": __host_threat_overview(host_threats),
            "severities": __host_severity_overview(host_severities),
            "equipment": equipment,
            "results": host_results,
        }

    # lists with just one element can be parsed as dict by xmltodict
    if isinstance(results, dict):
        per_result(results)
    else:
        for result in results:
            per_result(result)
    threat_count_dict = {
        __threats[i]: count for i, count in enumerate(threat_count)
    }
    # sort by amount descending
    host_threat_count = dict(
        sorted(
            host_threat_count.items(),
            key=lambda x: sum(x[1].values()),
            reverse=True,
        )
    )

    return list(by_host.values()), host_threat_count, threat_count_dict


@measure_time
def transform(data: Dict[str, str]) -> Report:
    """
    transform will use the given dict of a scanreport from gvmd
    to create an easy to use data representation for visual reports.
    """
    if not data:
        raise ValueError("Need data to process")
    report = data.get("report")
    # sometimes gvmd reports have .report.report sometimes just .report
    report = report.get("report", report)

    task = report.get("task") or {}
    logger.info("data transformation")
    results, host_counts, nvts_counts = __create_results_per_host(report)

    return Report(
        report.get("id"),
        task.get("name"),
        task.get("comment"),
        report.get("scan_start"),
        Overview(
            hosts=host_counts, nvts=nvts_counts, vulnerable_equipment=None
        ),
        results,
    )



# -*- coding: utf-8 -*-
# Copyright (C) 2020-2021 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
from typing import Callable, Tuple, Union

from rest_framework.authentication import BaseAuthentication
from rest_framework import exceptions
from rest_framework.request import HttpRequest

import requests
from requests.models import Response
import xmltodict

from pheme import settings


logger: logging.Logger = logging.getLogger(__name__)

# pylint falsely identifies Union as unsubscriptable-object in python 3.9
# pylint: disable=E1136


def __gsad_user_role(
    token: str, gsad_sid: str, *, gsad_url: str, get: Callable
) -> Tuple[Union[str, None], Union[str, None]]:
    if not gsad_url:
        logger.warning("no gsad url")
        return None, None
    params = {"token": token, "cmd": "get_users"}
    cookies = dict(GSAD_SID=gsad_sid)
    response: Response = get(gsad_url, params=params, cookies=cookies)
    resp = xmltodict.parse(
        response.text, attr_prefix="", cdata_key="text", dict_constructor=dict
    ).get("envelope")
    return resp.get("login"), resp.get("role")


def get_username_role(
    request: HttpRequest,
    *,
    gsad_url: str = settings.GSAD_URL,
    get: Callable = requests.get,
) -> Tuple[Union[str, None], Union[str, None]]:
    """
    returns a username and a role when it got found based on the request.

    So far only gsad is supported.
    Parameter:
        request: HttpRequest, the incoming request is used to get information
        gsad_url: str, when using gsad the url of gsad must be set, default is
            pheme.settings.GSAD_URL
        get: Callable, a lambda to get the user and tole. It must return a
            requests.models.Response

    Returns:
        A tuple either None, None or user, role
    """
    gsad_sid = request.COOKIES.get("GSAD_SID")
    token = request.query_params.get("token")
    if token and gsad_sid:
        return __gsad_user_role(token, gsad_sid, gsad_url=gsad_url, get=get)
    return None, None


class SimpleApiKeyAuthentication(BaseAuthentication):
    def authenticate(self, request: HttpRequest):
        api_key = request.META.get("HTTP_X_API_KEY", "")
        if api_key != settings.SECRET_KEY:
            raise exceptions.AuthenticationFailed("Invalid api-key.")


class LoggedInAsAUser(BaseAuthentication):
    def authenticate(self, request: HttpRequest):
        username, role = get_username_role(request)
        if username is None:
            return None
        request.META["GVM_USERNAME"] = username
        request.META["GVM_ROLE"] = username
        return (username, role)



# -*- coding: utf-8 -*-
# pheme/templatetags/dynamic_template.py
# Copyright (C) 2020-2021 Greenbone AG
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Adds the possibility to print iso_8601 strings as %a, %b %d, %Y %I %p %Z', e.g.:
    2001-07-22T09:15:37Z -> Sun, Jul 22, 2001 09 AM UTC
"""
from datetime import datetime
from django.utils.safestring import SafeText, SafeString
from django import template


register = template.Library()


@register.filter
def format_time(iso_8601: str) -> SafeString:
    """
    Transforms iso string to %a, %b %d, %Y %I %p %Z.

    Parameter:
        iso_8601 - the string to format

    Returns:
        An the input parameter on failure or an transformed SafeString with
        human readable time format.
    """

    try:
        time = datetime.strptime(iso_8601, "%Y-%m-%dT%H:%M:%S%z")
        return SafeText(datetime.strftime(time, "%a, %b %d, %Y %I %p %Z"))
    except (ValueError, TypeError) as _:
        return SafeText(iso_8601)



# -*- coding: utf-8 -*-
# pheme/renderer/xml.py
# Copyright (C) 2020-2021 Greenbone AG
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import List, Dict, Generator, Union
from io import StringIO
from csv import DictWriter
from rest_framework.renderers import BaseRenderer
import xmltodict


class CSVRenderer(BaseRenderer):
    media_type = "text/csv"
    format = "text"
    charset = "utf-8"

    def __flatten_per_result(
        self, data: Dict
    ) -> Generator[Union[List[str], Dict], None, None]:
        # for the case that results is there but it is None
        results = data.pop("results", None) or []
        send_keys = True
        for host in results:
            for result in host.get("results"):
                flatten = {
                    **data,
                    "os": host.get("os"),
                    **result.pop("nvt_tags_interpreted", {}),
                    **result,
                }
                if send_keys:
                    yield flatten.keys()
                    send_keys = False
                yield flatten

    def render(self, data, accepted_media_type=None, renderer_context=None):
        if data is None:
            return ""
        try:
            generator = self.__flatten_per_result(data)
            result = StringIO()
            writer = DictWriter(result, next(generator))
            writer.writeheader()
            writer.writerows(generator)
            result.seek(0)
            return result.read()
        except StopIteration:
            return ""


class MarkDownTableRenderer(BaseRenderer):
    media_type = "text/markdown+table"
    format = "text"
    charset = "utf-8"

    def __as_md(self, previous_key, data):
        def append(value):
            return isinstance(value, str) or isinstance(value, int)

        if not (isinstance(data, dict) or isinstance(data, list)):
            return []
        items = None
        result = []

        def to_new_key(key: str):
            if not previous_key:
                return key
            if not key:
                return previous_key
            return f"{previous_key}.{key}"

        if isinstance(data, list):
            result.append(
                f"|{previous_key}|{{% for item in {previous_key} %}} ..."
                " {% endfor %}| a list of items|"
            )
            previous_key = "item"
            items = [("", v) for v in data]
        else:
            items = data.items()
        for key, value in items:
            new_key = to_new_key(key)
            if append(value):
                result.append(f"|{new_key}|{{ {new_key} }}|{value}|")
            else:
                result = result + self.__as_md(new_key, value)
        return result

    def render(self, data, accepted_media_type=None, renderer_context=None):
        if data is None:
            return ""
        first_line = "|key|template_example|description|\n"
        table_indicator = "| :- | :--: | -: |\n"
        rest = "\n".join(self.__as_md("", data))
        return first_line + table_indicator + rest


class XMLRenderer(BaseRenderer):
    media_type = "application/xml"
    format = "xml"
    charset = "utf-8"

    def render(self, data, accepted_media_type=None, renderer_context=None):
        if data is None:
            return ""

        return xmltodict.unparse(data)



# -*- coding: utf-8 -*-
# Copyright (C) 2020-2021 Greenbone AG
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# Squarified Treemap Layout
# Implements algorithm from Bruls, Huizing, van Wijk, "Squarified Treemaps"
#   (but not using their pseudocode)

# It is a simplified, adapted version of squarify:
# https://github.com/laserson/squarify
# For more information see:
# https://www.win.tue.nl/~vanwijk/stm.pdf
# pylint: disable=C0103
import math as m
import numbers
from dataclasses import dataclass
from typing import Dict, List

from django.utils.safestring import SafeText

from pheme.templatetags.charts import (
    _severity_class_colors,
    register,
    build_legend,
    calculate_legend_start_height,
)

__ELEMENT_TEMPLATE = """
<g>
    <rect x="{x}" y="{y}" width="{width}" height="{height}" fill="{color}" stroke="{border_color}" strokeWidth="1" />
    <text style="font-size:{font_size}px;font-family:{font_family}" x="{label_x}" y="{label_y}" width="{width}" height="{height}" dominant-baseline="central">{label}</text>
</g>
"""

__TEMPLATE = """
<svg version="1.1"
     baseProfile="full"
     width="{width}" height="{height}"
     viewBox="0 0 {width} {height}"
     xmlns="http://www.w3.org/2000/svg">
     <g transform="translate({start_tree}, 0)">
     {rects}
     </g>
     {legend}
</svg>
"""


@dataclass
class Rect:
    x: float
    y: float
    dx: float
    dy: float


def __create_rectangle(x: float, y: float, dx: float, dy: float) -> Rect:
    # add spacing of 1px
    if dx > 2:
        x += 1
        dx -= 2
    if dy > 2:
        y += 1
        dy -= 2
    return Rect(x, y, dx, dy)


def __layoutrow(sizes: List[numbers.Number], rect: Rect) -> List[Rect]:
    covered_area = sum(sizes)
    width = covered_area / rect.dy
    rects = []
    y = rect.y
    for size in sizes:
        rects.append(__create_rectangle(rect.x, y, width, size / width))
        y += size / width
    return rects


def __layoutcol(sizes: List[numbers.Number], rect: Rect) -> List[Rect]:
    covered_area = sum(sizes)
    height = covered_area / rect.dx
    rects = []
    x = rect.x
    for size in sizes:
        rects.append(__create_rectangle(x, rect.y, size / height, height))
        x += size / height
    return rects


def __layout(sizes: List[numbers.Number], rect: Rect) -> List[Rect]:
    if rect.dx >= rect.dy:
        return __layoutrow(sizes, rect)
    return __layoutcol(sizes, rect)


def __leftoverrow(covered_area: float, rect: Rect) -> Rect:
    width = covered_area / rect.dy
    return Rect(rect.x + width, rect.y, rect.dx - width, rect.dy)


def __leftovercol(covered_area: float, rect: Rect) -> Rect:
    height = covered_area / rect.dx
    return Rect(rect.x, rect.y + height, rect.dx, rect.dy - height)


def __leftover(sizes: List[numbers.Number], rect: Rect) -> Rect:
    covered_area = sum(sizes)
    if rect.dx >= rect.dy:
        return __leftoverrow(covered_area, rect)
    return __leftovercol(covered_area, rect)


def __find_split(sizes: List[numbers.Number], rect: Rect) -> int:
    """
    returns the index to split the sizes based on worst ratio to get the
    remaining and current space.

    For an example the area ratio of 5, 3, 2, 1 with the area of 90 * 50
    >>> test_data = [2045.5, 1227.3, 818.2, 409.0 ]
    >>> __find_split(test_data, Rect(0, 0, 90, 50))
    1

    The first two elements can be put into the first space and the rest needs to
    calculated for the remaining space.

    """

    def worst(i: int) -> float:
        return max(
            [
                max(rect.dx / rect.dy, rect.dy / rect.dx)
                for rect in __layout(sizes[:i], rect)
            ]
        )

    for i in range(1, len(sizes)):
        if worst(i) < worst(i + 1):
            return i
    return len(sizes) - 1


def __squarify(sizes: List[numbers.Number], rect: Rect) -> List[Rect]:
    """
    calculates treemap rectangles using an algorithm based on Bruls, Huizing,
    van Wijk, "Squarified Treemaps" and "squarify":
    https://github.com/laserson/squarify

    >>> __squarify([5, 1], Rect(0, 0, 90, 50))
    [Rect(x=1, y=1, dx=73.0, dy=48.0), Rect(x=76.0, y=1, dx=13.0, dy=48.0)]

    Parameters:
        sizes : list-like of numeric values
            The set of values to compute a treemap for. `sizes` must be sorted,
            positive values
        x, y : numeric
            The coordinates of the "origin".
        dx, dy : numeric
            The full width (`dx`) and height (`dy`) of the treemap.

    Returns:
        List[Rect]
            Each dict in the returned list represents a single rectangle in the
            treemap. The order corresponds to the input order.
    """
    if len(sizes) == 0:
        return []

    total_size = sum(sizes)
    total_area = rect.dx * rect.dy

    sizes = list([size * total_area / total_size for size in sizes])

    if len(sizes) == 1:
        return __layout(sizes, rect)

    i = __find_split(sizes, rect)
    current = sizes[:i]
    remaining = sizes[i:]

    return __layout(current, rect) + __squarify(
        remaining, __leftover(current, rect)
    )


def __transform_to_tree_data(data) -> List[Dict]:
    """
    tansforms given data to treemap compatible format.
    The support types are:
    >>> example_host_data = dict(
    ...     host1=dict(high=12, medium=4, low=0),
    ...     host2=dict(high=0, medium=4, low=0),
    ... )
    >>> __transform_to_tree_data(example_host_data)
    ([16, 4], ['host1', 'host2'], ['high', 'medium'])

    On unsupported input data it returns empty lists:
    >>> __transform_to_tree_data(dict(k="hello", v="world"))
    ([], [], [])
    >>> __transform_to_tree_data(12)
    ([], [], [])
    """
    values = []
    labels = []
    color_keys = []
    if isinstance(data, dict):
        for key, val in data.items():
            color = None
            val_sum = 0
            if isinstance(val, dict):
                for item_key, item_val in val.items():
                    if isinstance(item_val, numbers.Number):
                        val_sum += item_val
                        if color is None and item_val > 0:
                            color = item_key
            if color is not None and val_sum > 0:
                values.append(val_sum)
                labels.append(key)
                color_keys.append(color)
    return values, labels, color_keys


@register.filter
def treemap(
    data: List[Dict],
    width: int = 1024,
    height: int = 768,
    border_color: str = "#ffffff",
    title_color: Dict[str, str] = None,
    font_size: int = 10,
    font_family: str = "Dejavu Sans",
) -> SafeText:
    """
    Expects a sorted dict containing a str, and a dict with values in it.

    An example can be:

    {
        'host1': {'high': 12, 'medium': 4, 'low': 0},
        'host2': {'high': 0, 'medium': 4, 'low': 0},
    }

    With that host1 and host2 will be used as a label, high and medium will be
    used to find the color and values will be summed for the rectangle
    calculation.

    Data will be transformed to lists of labels, color_keys and the actual
    values.

    The color key within the data  must be consistent with title_color.

    Parameters:
        data: needs to be an dict containing label, color_key and values.
        width: width of the svg (default 1024)
        height: height of the svg (default 768)
        border_color: color of the rectangle border (default white)
        title_color: the color_key to color lookup map
            (default _severity_class_colors)
        font_family - the font family used within text elements
        font_size - the font size used within text elements

    Returns:
        the treemap in svg as a SafeString.
    """
    if not title_color:
        title_color = _severity_class_colors
    sizes, label, color_keys = __transform_to_tree_data(data)
    max_legend_len = max([len(k) for k in title_color.keys()])
    s_width = width - max_legend_len * font_size - font_size
    sizes = __squarify(sizes, Rect(0, 0, s_width, height))
    elements = ""
    for i, d in enumerate(sizes):
        label_size_in_px = len(label[i]) * font_size
        label_x = d.x + 1
        # move half of the font size down and add a buffer for different
        # heights of characters (depends on font) to display the label
        # on the upper left corner of an rectangle.
        label_y = d.y + font_size / 2 + 5
        max_label_len = m.ceil(d.dx / (label_size_in_px * 0.75) * len(label[i]))
        if d.dy <= font_size:
            max_label_len = 0

        elements += __ELEMENT_TEMPLATE.format(
            x=d.x,
            y=d.y,
            width=d.dx,
            height=d.dy,
            color=title_color.get(color_keys[i]),
            border_color=border_color,
            label_x=label_x,
            label_y=label_y,
            label=label[i][:max_label_len],
            font_size=font_size,
            font_family=font_family,
        )
    legend_start = calculate_legend_start_height(height, title_color, font_size)
    return SafeText(
        __TEMPLATE.format(
            width=width,
            height=height,
            rects=elements,
            legend=build_legend(legend_start, title_color),
            start_tree=width - s_width,
        )
    )



# -*- coding: utf-8 -*-
# pheme/templatetags/dynamic_template.py
# Copyright (C) 2020-2021 Greenbone AG
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Adds the possibility to dynamially render template snippets load via parameter
and render them into another template without having the need to declare them
as a dependency upfront.
"""

from django.utils.safestring import mark_safe, SafeString
from django import template
from pheme.parameter import load_params


register = template.Library()


@register.filter
def dynamic_template(data, key: str) -> SafeString:
    """
    Loads the template via key and renders the template with the given
    data.

    Parameter:
        data - used to create the context for the renderer
        key: str - identifier of the template to render

    Returns:
        The rendered template or an empty string when the key is not found
    """
    tmpl = load_params().get(key)
    if not tmpl:
        return mark_safe("")
    return mark_safe(template.Template(tmpl).render(template.Context(data)))



# -*- coding: utf-8 -*-
# Copyright (C) 2020-2021 Greenbone AG
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Dict
from django import template

_severity_class_colors = {
    "High": "#d4003e",
    "Medium": "#fcb900",
    "Low": "#7db4d0",
}

register = template.Library()

__LEGEND_TEMPLATE = """
<g transform="translate({x}, {y})">
{legend}
</g>
"""
__LEGEND_ELEMENT = """
<rect x="{x}" y="{y}" height="{font_size}" width="{font_size}" style="fill: {color};"></rect>
<text style="font-size:{font_size}px;font-family:{font_family}" x="{text_x}" y="{text_y}" dominant-baseline="central">{label}</text>
"""


def calculate_legend_start_height(
    height: int, label_color: Dict, font_size: int
) -> int:
    return int(height / 2 - (len(label_color) * font_size) / 2)


def build_legend(
    start_height: int,
    label_color: Dict,
    font_family: str = "Dejavu Sans",
    font_size: int = 10,
) -> str:
    """
    Generates a legend at     start_height y position and at
    width / 2 - last_x of element / 2 x position.

    By iterating through the the label_color dict and using the key as a label
    and the value as a fill color for the rectangle.

    Parameters:
        start_height: x position of legend
        label_color: key is used as a label and the value as a fill color for an
            rectangle (font_size * font_size).
        font_family - the font family used within text elements
        font_size - the font size used within text elements
    Returns:
        A complete g element with legends as a string.

    """
    legend_elements = ""
    x_pos = 0
    y_pos = 0
    # offset of two to the rect element
    text_x_pos = font_size + 2
    for label, color in label_color.items():
        legend_elements += __LEGEND_ELEMENT.format(
            x=x_pos,
            y=y_pos,
            font_size=font_size,
            font_family=font_family,
            color=color,
            label=label,
            text_x=text_x_pos,
            # text position needs to aligned in the middle of the rect
            text_y=y_pos + font_size / 2,
        )
        y_pos += font_size + font_size / 2

    return __LEGEND_TEMPLATE.format(x=0, y=start_height, legend=legend_elements)



# -*- coding: utf-8 -*-
# pheme/parser/xml.py
# Copyright (C) 2020-2021 Greenbone AG
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Provides XML parsing support.

Although there is the project djangorestframework-xml xmltodict seems to be more
reliable for our usecase.

"""
from rest_framework.parsers import BaseParser
import xmltodict


class XMLFormParser(BaseParser):
    """
    XML parser based on xmltodict.
    """

    media_type = "multipart/form-data"

    def parse(self, stream, media_type=None, parser_context=None):
        for report in stream.FILES.values():
            if report.content_type == "text/xml":
                return xmltodict.parse(
                    report.read(),
                    attr_prefix="",
                    cdata_key="text",
                    dict_constructor=dict,
                )
        return None


class XMLParser(BaseParser):
    """
    XML parser based on xmltodict.
    """

    media_type = "application/xml"

    def parse(self, stream, media_type=None, parser_context=None):
        return xmltodict.parse(
            stream, attr_prefix="", cdata_key="text", dict_constructor=dict
        )



# -*- coding: utf-8 -*-
# pheme/transformation/scanreport/model.py
# Copyright (C) 2020-2021 Greenbone AG
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# pylint: disable=W0614,W0511,W0401,C0103
from dataclasses import dataclass
from typing import Dict, List, Union


@dataclass
class NVTThreatCount:
    level: str
    count: int


@dataclass
class Equipment:
    os: str  # search for host.text and refs than os-detection
    ports: List[str]  # ports port host and ports port text


@dataclass
class NVTCount:
    oid: str
    amount: int
    name: str


@dataclass
class HostCount:
    ip: str
    amount: int
    name: str


@dataclass
class CountGraph:
    name: str
    chart: str
    counts: List[Union[NVTCount, HostCount]]


@dataclass
class SeverityCount:
    severity: str
    amount: int


@dataclass
class Overview:
    hosts: CountGraph
    nvts: CountGraph
    vulnerable_equipment: CountGraph


@dataclass
class Report:
    id: str
    name: str
    comment: str
    start: str
    overview: Overview
    results: List[Dict]


def describe():
    return Report(
        id="str; identifier of a report",
        name="str; name of the scan",
        comment="str; comment of the scan",
        start="str; datetime of a scan start",
        overview=Overview(
            hosts=CountGraph(
                name="host_top_ten",
                chart="str; link to chart image (base64 encoded datalink)",
                counts=[],
            ),
            nvts=CountGraph(
                name="nvt_overview",
                chart="str; link to chart image (base64 encoded datalink)",
                counts=[],
            ),
            vulnerable_equipment=CountGraph(
                name="vulnerable_equipment",
                chart="str; link to chart image (base64 encoded datalink)",
                counts=[],
            ),
        ),
        results=[
            dict(
                host="str; ip address of host",
                threats={
                    "high": "int; amount of high nvts in host",
                    "medium": "int; amount of medium nvts in host",
                    "low": "int; amount of low nvts in host",
                    "total": "int; amount of nvts in host",
                    "highest": "str; highest threat within host",
                },
                severities={
                    "1": "int; amount of 1 severities in host",
                    "2": "int; amount of 2 severities in host",
                    "3": "int; amount of 3 severities in host",
                    "4": "int; amount of 4 severities in host",
                    "5": "int; amount of 5 severities in host",
                    "6": "int; amount of 6 severities in host",
                    "7": "int; amount of 7 severities in host",
                    "8": "int; amount of 8 severities in host",
                    "9": "int; amount of 9 severities in host",
                    "10": "int; amount of 10 severities in host",
                    "total": "int; amount of nvts in host",
                    "highest": "str; highest severity within host",
                },
                equipment=Equipment(
                    os="str; operating system", ports=["str; open ports"]
                ),
                results={
                    "nvt.oid": "str; nvt.oid; optional",
                    "nvt.type": "str; nvt.type; optional",
                    "nvt.name": "str; nvt.name; optional",
                    "nvt.family": "str; nvt.family; optional",
                    "nvt.cvss_base": "str; nvt.cvss_base; optional",
                    "nvt.tags_interpreted": {
                        "name_of_tag": "str; value of a tag ()",
                        "may_contain_multiple_tags": "str; depends on xml",
                        "solution": "str; solution the fix the problem",
                        "solution_type": "str; solution type",
                        "insight": "str; insight into the problem",
                        "vuldetect": "str; ",
                        "cvss_base_vector": "str, cvss base vector",
                    },
                    "nvt.refs.ref": [{"name_of_ref": ["str; ref_vallue"]}],
                    "nvt.solution.type": "str; nvt.solution.type; optional",
                    "nvt.solution.text": "str; nvt.solution.text; optional",
                    "port": "str; port; optional",
                    "severity": "str; severity; optional",
                    "qod.value": "str; qod.value; optional",
                    "qod.type": "str; qod.type; optional",
                    "description": "str; description; optional",
                },
            )
        ],
    )
